# -*- coding: utf-8 -*-

from rest_framework.exceptions import ValidationError
from rest_framework.filters import BaseFilterBackend
from rest_framework.pagination import LimitOffsetPagination
from rest_framework.views import APIView

from rest_search.schemas import get_form_schema


class SearchFilterBackend(BaseFilterBackend):
    """
    Dummy filter backend to enable API documentation.
    """
    def get_schema_fields(self, view):
        return get_form_schema(view.form_class)


class SearchAPIView(APIView):
    filter_backends = (SearchFilterBackend,)

    def get(self, request, *args, **kwargs):
        query = self.get_query()
        sort = self.get_sort()

        pagination = LimitOffsetPagination()
        pagination.default_limit = 20
        pagination.limit = pagination.get_limit(request)
        pagination.offset = pagination.get_offset(request)
        pagination.request = request

        body = {
            'query': query,
            'size': pagination.limit,
            'from': pagination.offset
        }
        if sort:
            body['sort'] = sort

        # execute elasticsearch query
        indexer = self.get_indexer()
        res = indexer.search(body=body)

        # map back to expected format
        items = list(indexer.map_results(res['hits']['hits']))
        pagination.count = res['hits']['total']
        return pagination.get_paginated_response(items)

    def get_indexer(self):
        return self.indexer_class()

    def get_query(self):
        """
        Returns the 'query' element of the ElasticSearch request body.
        """
        form = self.form_class(self.request.GET, context={
            'request': self.request
        })
        if not form.is_valid():
            raise ValidationError(form.errors)
        return form.get_query()

    def get_sort(self):
        """
        Returns the 'sort' element of the ElastiSearch request body.

        To use the default sorting (by score) return None.
        """
        return None



# -*- coding: utf-8 -*-

import coreapi
import coreschema
from django import forms
from django.utils.encoding import force_text


def get_form_schema(form_class):
    """
    Return the coreapi schema for the given form class.
    """
    fields = []
    for name, field in form_class.declared_fields.items():
        fields.append(coreapi.Field(
            location='query',
            name=name,
            required=field.required,
            schema=get_form_field_schema(field)))
    return fields


def get_form_field_schema(field):
    """
    Returns the coreapi schema for the given form field.
    """
    title = force_text(field.label) if field.label else ''
    description = force_text(field.help_text) if field.help_text else ''

    if isinstance(field, forms.BooleanField):
        field_class = coreschema.Boolean
    elif isinstance(field, forms.DateTimeField):
        return coreschema.String(description=description, title=title, format='date-time')
    elif isinstance(field, forms.FloatField):
        field_class = coreschema.Number
    elif isinstance(field, (forms.IntegerField, forms.ModelChoiceField)):
        field_class = coreschema.Integer
    else:
        field_class = coreschema.String

    return field_class(description=description, title=title)



# -*- coding: utf-8 -*-

from django.db.models.signals import post_delete, post_save
from elasticsearch.helpers import scan

from rest_search import get_elasticsearch


_REGISTERED_CLASSES = []


def bulk_iterate(queryset, block_size=1000):
    """
    Iterates over a huge queryset in chunks of 1000 items.
    """
    total = queryset.count()
    queryset = queryset.order_by('pk')
    for start in range(0, total, block_size):
        for item in queryset[start:start + block_size]:
            yield item


class Indexer(object):
    mappings = None
    private_properties = []

    def __init__(self):
        self.doc_type = self.serializer_class.Meta.model.__name__

    def map_results(self, results):
        """
        Removes ES-specific fields from results.
        """
        def map_result_item(x):
            item = x['_source']
            for key in self.private_properties:
                item.pop(key, None)
            return item
        return map(map_result_item, results)

    def scan(self, **kwargs):
        es = get_elasticsearch(self)
        return scan(es, index=es._index, doc_type=self.doc_type, **kwargs)

    def search(self, **kwargs):
        es = get_elasticsearch(self)
        return es.search(index=es._index, doc_type=self.doc_type, **kwargs)


def _get_registered():
    """
    Returns instances of all registered indexers.
    """
    return [cls() for cls in _REGISTERED_CLASSES]


def _instance_changed(sender, instance, **kwargs):
    """
    Queues an update to the ElasticSearch index.
    """
    from rest_search import queue_add
    doc_type = instance.__class__.__name__
    queue_add({
        doc_type: [instance.pk],
    })


def register(indexer_class):
    """
    Registers an indexer class.
    """
    if indexer_class not in _REGISTERED_CLASSES:
        _REGISTERED_CLASSES.append(indexer_class)

        # register signal handlers
        model = indexer_class.serializer_class.Meta.model
        post_delete.connect(_instance_changed, sender=model)
        post_save.connect(_instance_changed, sender=model)



# -*- coding: utf-8 -*-

import logging

from celery import shared_task
from elasticsearch.helpers import bulk

from rest_search import get_elasticsearch
from rest_search.indexers import bulk_iterate, _get_registered


logger = logging.getLogger('rest_search')


def create_index():
    """
    Creates the ElasticSearch index if it does not exist.
    """
    conns = {}
    for indexer in _get_registered():
        es = get_elasticsearch(indexer)
        if es not in conns:
            conns[es] = {
                'mappings': {},
                'settings': es._settings
            }
        if indexer.mappings is not None:
            mappings = conns[es]['mappings']
            mappings[indexer.doc_type] = indexer.mappings

    for es, body in conns.items():
        if not es.indices.exists(es._index):
            es.indices.create(index=es._index, body=body)


def delete_index(es):
    """
    Deletes the ElasticSearch index.
    """
    return es.indices.delete(index=es._index, ignore=404)


@shared_task
def patch_index(updates):
    """
    Performs a partial update of the ElasticSearch index.
    """
    updates_str = ['%s: %d items' % (k, len(v)) for k, v in updates.items()]
    logger.info('Patching index (%s)' % ', '.join(updates_str))

    indexers = _get_registered()
    for doc_type, pks in updates.items():
        for indexer in indexers:
            if indexer.doc_type == doc_type:
                _patch_index(indexer, pks)


@shared_task
def update_index(remove=True):
    """
    Performs a full update of the ElasticSearch index.
    """
    logger.info('Updating index')

    create_index()

    for indexer in _get_registered():
        _update_index(indexer, remove=remove)


def _delete_items(indexer, pks):
    es = get_elasticsearch(indexer)

    def mapper(pk):
        return {
            '_index': es._index,
            '_type': indexer.doc_type,
            '_id': pk,
            '_op_type': 'delete',
        }

    bulk(es, map(mapper, pks), raise_on_error=False)


def _index_items(indexer, queryset):
    es = get_elasticsearch(indexer)
    seen_pks = set()

    def mapper(item):
        seen_pks.add(item.pk)
        return {
            '_index': es._index,
            '_type': indexer.doc_type,
            '_id': item.pk,
            '_source': indexer.serializer_class(item).data
        }

    bulk(es, map(mapper, bulk_iterate(queryset)))
    return seen_pks


def _patch_index(indexer, pks):
    # index current items
    queryset = indexer.get_queryset().filter(pk__in=pks)
    seen_pks = _index_items(indexer, queryset)

    # remove obsolete items
    removed_pks = set(pks) - seen_pks
    _delete_items(indexer, removed_pks)


def _update_index(indexer, remove):
    scan = indexer.scan(query={'stored_fields': []})
    old_pks = set([int(i['_id']) for i in scan])

    # index current items
    queryset = indexer.get_queryset()
    seen_pks = _index_items(indexer, queryset)

    # remove obsolete items
    if remove:
        removed_pks = old_pks - seen_pks
        _delete_items(indexer, removed_pks)



# -*- coding: utf-8 -*-

from threading import local

import certifi
from aws_requests_auth.aws_auth import AWSRequestsAuth
from django.conf import settings
from elasticsearch import Elasticsearch, RequestsHttpConnection


DEFAULT_INDEX_SETTINGS = {
    'analysis': {
        'analyzer': {
            'default': {
                'tokenizer': 'standard',
                'filter': [
                    'standard',
                    'lowercase',
                    'asciifolding',
                ]
            }
        }
    }
}
QUEUES = {}


class ConnectionHandler(object):
    def __init__(self):
        self._connections = local()

    def __delitem__(self, alias):
        delattr(self._connections, alias)

    def __getitem__(self, alias):
        if hasattr(self._connections, alias):
            return getattr(self._connections, alias)

        conn = self.__create_connection(
            settings.REST_SEARCH_CONNECTIONS[alias])
        setattr(self._connections, alias, conn)
        return conn

    def __create_connection(self, config):
        kwargs = {
            'host': config['HOST'],
            'port': config.get('PORT', 9200),
            'use_ssl': config.get('USE_SSL', False),
            'verify_certs': True,
            'ca_certs': certifi.where()
        }

        if 'AWS_ACCESS_KEY' in config and \
           'AWS_SECRET_KEY' in config and \
           'AWS_REGION' in config:

            kwargs['connection_class'] = RequestsHttpConnection
            kwargs['http_auth'] = AWSRequestsAuth(
                aws_access_key=config['AWS_ACCESS_KEY'],
                aws_secret_access_key=config['AWS_SECRET_KEY'],
                aws_host=config['HOST'],
                aws_region=config['AWS_REGION'],
                aws_service='es')

        es = Elasticsearch(**kwargs)
        es._index = config['INDEX_NAME']
        es._settings = config.get('INDEX_SETTINGS', DEFAULT_INDEX_SETTINGS)
        return es


def get_elasticsearch(indexer=None):
    return connections['default']


def queue_add(updates):
    """
    Adds items to the updates queue, in the form:

    {
        'Book': [1, 2],
    }
    """
    if getattr(settings, 'SEARCH_UPDATES_ENABLED', True):
        for doc_type, pks in updates.items():
            if doc_type in QUEUES:
                QUEUES[doc_type].update(set(pks))
            else:
                QUEUES[doc_type] = set(pks)


def queue_flush():
    """
    Triggers a celery task for the queued updates, if any.
    """
    global QUEUES
    if QUEUES:
        from rest_search.tasks import patch_index
        # convert sets to lists, otherwise they are not JSON-serializable
        args = {}
        for doc_type, pks in QUEUES.items():
            args[doc_type] = list(pks)
        patch_index.delay(args)
        QUEUES.clear()


connections = ConnectionHandler()



# -*- coding: utf-8 -*-

from django import forms


class SearchMixin(object):
    """
    Mixin to help build ElasticSearch queries.
    """
    def get_query(self):
        """
        Returns the query to be executed by ElasticSearch.
        """
        return self._score_query(self._build_query())

    def _build_query(self):
        query = {
            'bool': {}
        }

        for kind in ['filter', 'must', 'must_not', 'should']:
            clauses = getattr(self, 'get_%s_clauses' % kind)()
            if clauses:
                query['bool'][kind] = clauses

        if query['bool']:
            return query
        else:
            return {
                'match_all': {}
            }

    def get_filter_clauses(self):
        return []

    def get_must_clauses(self):
        return []

    def get_must_not_clauses(self):
        return []

    def get_should_clauses(self):
        return []

    def _score_query(self, query):
        return query


class SearchForm(SearchMixin, forms.Form):
    """
    Base form for building ElasticSearch queries.
    """
    def __init__(self, data, context={}):
        self.context = context
        super(SearchForm, self).__init__(data)
